package genius.core.boaframework;

import java.io.Serializable;
import java.util.Map;

import genius.core.Bid;
import genius.core.issue.Issue;
import genius.core.protocol.BilateralAtomicNegotiationSession;
import genius.core.utility.AbstractUtilitySpace;
import genius.core.utility.AdditiveUtilitySpace;

/**
 * Describes an opponent model of an agent of the BOA framework. This model
 * assumes issue weights hence only supports {@link AdditiveUtilitySpace}.
 * 
 * Tim Baarslag, Koen Hindriks, Mark Hendrikx, Alex Dirkzwager and Catholijn M.
 * Jonker. Decoupling Negotiating Agents to Explore the Space of Negotiation
 * Strategies
 * 
 * @author Alex Dirkzwager, Mark Hendrikx
 */
public abstract class OpponentModel extends BOA {

	/** Reference to the estimated opponent's utility state */
	protected AdditiveUtilitySpace opponentUtilitySpace;
	/** Boolean to indicate that the model has been cleared to free resources */
	private boolean cleared;

	/**
	 * Initializes the model. The init method should always be called after
	 * creating an opponent model.
	 * 
	 * @param negotiationSession
	 *            reference to the state of the negotiation
	 * @param parameters
	 * @throws Exception
	 */
	public void init(NegotiationSession negotiationSession, Map<String, Double> parameters) {
		super.init(negotiationSession, parameters);
		opponentUtilitySpace = (AdditiveUtilitySpace) negotiationSession.getUtilitySpace().copy();
		cleared = false;
	}

	/**
	 * Method used to receiveMessage the opponent model.
	 * 
	 * @param opponentBid
	 */
	public void updateModel(Bid opponentBid) {
		updateModel(opponentBid, negotiationSession.getTime());
	}

	/**
	 * Method used to receiveMessage the opponent model.
	 * 
	 * @param bid
	 *            to receiveMessage the model with.
	 * @param time
	 *            at which the bid was offered.
	 */
	public abstract void updateModel(Bid bid, double time);

	/**
	 * Determines the utility of a bid according to the preference profile.
	 * 
	 * @param bid
	 *            of which the utility is calculated.
	 * @return Utility of the bid
	 */
	public double getBidEvaluation(Bid bid) {
		try {
			return opponentUtilitySpace.getUtility(bid);
		} catch (Exception e) {
			e.printStackTrace();
		}
		return -1;
	}

	/**
	 * @return the estimated utility space of the opponent
	 */
	public AbstractUtilitySpace getOpponentUtilitySpace() {
		return opponentUtilitySpace;
	}

	/**
	 * Method which may be overwritten by an opponent model to get access to the
	 * opponent's utilityspace.
	 * 
	 * @param fNegotiation
	 */
	public void setOpponentUtilitySpace(BilateralAtomicNegotiationSession fNegotiation) {
	}

	/**
	 * Method which may be overwritten by an opponent model to get access to the
	 * opponent's utilityspace.
	 * 
	 * @param opponentUtilitySpace
	 */
	public void setOpponentUtilitySpace(AdditiveUtilitySpace opponentUtilitySpace) {
	}

	/**
	 * Returns the weight of a particular issue in the domain. Only works with
	 * {@link AdditiveUtilitySpace}.
	 * 
	 * @param issue
	 *            from which the weight should be returned
	 * @return weight of the given issue
	 */
	public double getWeight(Issue issue) {
		return ((AdditiveUtilitySpace) opponentUtilitySpace).getWeight(issue.getNumber());
	}

	/**
	 * @return set of all estimated issue weights.
	 */
	public double[] getIssueWeights() {
		double estimatedIssueWeights[] = new double[negotiationSession.getUtilitySpace().getDomain().getIssues()
				.size()];
		int i = 0;
		for (Issue issue : negotiationSession.getUtilitySpace().getDomain().getIssues()) {
			estimatedIssueWeights[i] = getWeight(issue);
			i++;
		}
		return estimatedIssueWeights;
	}

	/**
	 * Removes references to the objects used by the opponent model.
	 */
	public void cleanUp() {
		negotiationSession = null;
		cleared = true;
	}

	/**
	 * @return if the opponent model is in a usable state.
	 */
	public boolean isCleared() {
		return cleared;
	}

	/**
	 * @return name of the opponent model.
	 */
	public String getName() {
		return "Default";
	}

	@Override
	public final void storeData(Serializable object) {
		negotiationSession.setData(BoaType.OPPONENTMODEL, object);
	}

	@Override
	public final Serializable loadData() {
		return negotiationSession.getData(BoaType.OPPONENTMODEL);
	}

}
