package genius.core.uncertainty;

import java.util.ArrayList;
import java.util.List;

import genius.core.boaframework.SortedOutcomeSpace;
import genius.core.uncertainty.flattening.FlattenedUtilitySpace;
import genius.core.utility.AbstractUtilitySpace;
import genius.core.utility.AdditiveUtilitySpaceWithPerturbation;

/**
 * @author Dimitrios Tsimpoukis, April 2018
 * 
 * MAJOR Class for the introduction of uncertainty. It is used only in the core and keeps track of both the real preferences 
 * of the user as well as the perception of the agent. The preference profile that is given to the 
 * agents can be a "distorted utility space" in the cases that uncertainty is included in the form of a perturbation
 * to the real utilities (see {@link AdditiveUtilitySpaceWithPerturbation}}) or a flattened Utility Space (see {@link FlattenedUtilitySpace}), 
 * OR in the form of a Pairwise Comparison User Model (see {@link PairwiseComparisonUserModel}).
 *
 */

public class UncertainPreferenceContainer {
	
	private UNCERTAINTYTYPE type;
	private AbstractUtilitySpace realUtilitySpace;
	private PairwiseComparisonUserModel pairwiseCompUserModel;
	private AbstractUtilitySpace perceivedUtilitySpace;

	
	/**
	 * Constructor for a Perceived Utility Space user model
	 * 
	 * @param realUtilitySpace
	 * @param agentLoadingTYPE
	 * @param perceivedUtilitySpace
	 */
	
	public UncertainPreferenceContainer(AbstractUtilitySpace realUtilitySpace, UNCERTAINTYTYPE type, AbstractUtilitySpace perceivedUtilitySpace){
		this.realUtilitySpace = realUtilitySpace;
		this.perceivedUtilitySpace = perceivedUtilitySpace;
		this.type = type;
		this.pairwiseCompUserModel = null;
	}
	
	
	/**
	 * Constructor for the creation of a Pairwise Comparison User Model
	 * @param realUtilitySpace
	 * @param agentLoadingTYPE
	 * @param numberOfComparisons
	 */
	
	public UncertainPreferenceContainer(AbstractUtilitySpace realUtilitySpace, UNCERTAINTYTYPE type,
			int numberOfComparisons, double error, boolean experimental){
		this(realUtilitySpace, type, null);
		if (experimental == true) createExperimentalOutcomeComparisonUserModel(numberOfComparisons, error);
		else createOutcomeComparisonUserModel(numberOfComparisons, error);
	}

	public UncertainPreferenceContainer(AbstractUtilitySpace realUtilitySpace, UNCERTAINTYTYPE type){
		this(realUtilitySpace, type, null);
	}
	
	public UncertainPreferenceContainer(UncertainPreferenceContainer container){
		this.realUtilitySpace = container.realUtilitySpace;
		this.type = container.type;
		this.perceivedUtilitySpace = container.perceivedUtilitySpace;
		this.pairwiseCompUserModel = container.pairwiseCompUserModel;
	}
	
	/**
	 * Generates comparisons between different outcomes.
	 * At the moment each outcome is part of only one comparison.
	 * 
	 * 
	 * @param amount, amount of comparisons
	 */
	public void createOutcomeComparisonUserModel(int amount, double error) {
		SortedOutcomeSpace outcomeSpace = (new SortedOutcomeSpace(this.getRealUtilitySpace()));
		List<OutcomeComparison> bidComparisons = new ArrayList<OutcomeComparison>();
		bidComparisons = (new ComparisonGenerator(outcomeSpace)).generateComparisonsWithError(amount, error);
		this.pairwiseCompUserModel = new PairwiseComparisonUserModel(bidComparisons);
	}
	
	
	// TODO REMOVE LATER, ReFactor for copy-paste
	public void createExperimentalOutcomeComparisonUserModel(int amount, double error) {
		SortedOutcomeSpace outcomeSpace = (new SortedOutcomeSpace(this.getRealUtilitySpace()));
 		List<OutcomeComparison> bidComparisons = new ArrayList<OutcomeComparison>();
		bidComparisons = (new ComparisonGenerator(outcomeSpace)).generateComparisonsWithError(amount, error);
		this.pairwiseCompUserModel = new ExperimentalPairwiseComparisonUserModel(bidComparisons, realUtilitySpace);
	}



	public AbstractUtilitySpace getRealUtilitySpace() {
		return realUtilitySpace;
	}
	
	public PairwiseComparisonUserModel getPairwiseCompUserModel() {
		return pairwiseCompUserModel;
	}

	public AbstractUtilitySpace getPerceivedUtilitySpace() {
		return perceivedUtilitySpace;
	}


	public UNCERTAINTYTYPE getType() {
		return type;
	}
}

